import os
from pepeline import read, ImgFormat
import torch
from torch.utils.data import Dataset


class ImageDataset(Dataset):
    def __init__(self, image_dir, device, transform=None):
        self.image_dir = image_dir
        self.transform = transform
        self.image_files = os.listdir(image_dir)
        self.device = device

    def __len__(self):
        return len(self.image_files)

    def __getitem__(self, idx):
        img_path = os.path.join(self.image_dir, self.image_files[idx])
        image = read(img_path, img_format=ImgFormat.F32)
        image = torch.tensor(image, dtype=torch.float32, device=self.device).permute(
            2, 0, 1
        )
        if self.transform:
            image = self.transform(image).to(self.device)

        return image, self.image_files[idx]



from pepedp.scripts.archs.blocklines import calculate_image_blockiness
from pepedp.scripts.utils.objects import IQANode


class BlockinessThread(IQANode):
    def __init__(
        self,
        img_dir,
        batch_size: int = 8,
        thread: float = 0.5,
        median_thread=0,
        move_folder: str | None = None,
    ):
        super().__init__(
            img_dir, batch_size, thread, median_thread, move_folder, None, reverse=True
        )
        self.model = calculate_image_blockiness

    def forward(self, images):
        return self.model(images)



import os
import shutil

import torch
from torch.utils.data import DataLoader

from .module import ImageDataset
from tqdm import tqdm


class Thread:
    def __init__(self, name, thread):
        self.name = name
        self.thread = thread

    def __repr__(self):
        return f"Thread(Name = {self.name}, Thread = {self.thread}\n)"


class ThreadList:
    def __init__(self):
        self.mass = []

    def append(self, thread):
        self.mass.append(thread)

    def extend(self, thread_list):
        self.mass.extend(thread_list)

    def sort(self, reverse: bool = False):
        self.mass.sort(key=lambda item: item.thread, reverse=reverse)

    def __iter__(self):
        return iter(self.mass)

    def __getitem__(self, index):
        return self.mass[index]

    def __len__(self):
        return len(self.mass)


class IQANode:
    def __init__(
        self,
        img_dir,
        batch_size: int = 8,
        thread: float = 0.5,
        median_thread=0,
        move_folder: str | None = None,
        transform=None,
        reverse=False,
    ):
        self.device = (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
        dataset = ImageDataset(img_dir, self.device, transform)
        self.data_loader = DataLoader(dataset, batch_size=batch_size)
        self.img_dir: str = img_dir
        self.thread = thread
        self.reverse = reverse
        self.move_folder = move_folder
        if move_folder is not None:
            import os

            os.makedirs(move_folder, exist_ok=True)
        if median_thread:
            self.median_thread = median_thread
            self.thread_list = ThreadList()
        else:
            self.thread_list = None

    def forward(self, images):
        raise NotImplementedError(
            "The forward method must be implemented in a subclass"
        )

    @torch.inference_mode()
    def __call__(self):
        for images, filenames in tqdm(self.data_loader):
            iqa = self.forward(images)
            for index in range(iqa.shape[0]):
                file_name = filenames[index]
                iqa_value = iqa[index]

                if self.thread_list is None:
                    if iqa[index] > self.thread and self.move_folder:
                        shutil.move(  # shutil.clone(
                            os.path.join(self.img_dir, file_name),
                            os.path.join(self.move_folder, file_name),
                        )
                    elif iqa[index] < self.thread and not self.move_folder:
                        os.remove(os.path.join(self.img_dir, file_name))
                else:
                    if (iqa[index] > self.thread and not self.reverse) or (
                        iqa[index] < self.thread and self.reverse
                    ):
                        self.thread_list.append(
                            Thread(name=file_name, thread=float(iqa_value))
                        )
                    else:
                        if not self.move_folder:
                            os.remove(os.path.join(self.img_dir, file_name))
        if self.thread_list:
            self.thread_list.sort(self.reverse)
            clip_index = int(len(self.thread_list) * self.median_thread)
            if self.move_folder:
                for thread in self.thread_list[-clip_index:]:
                    file_name = thread.name
                    shutil.move(
                        os.path.join(self.img_dir, file_name),
                        os.path.join(self.move_folder, file_name),
                    )
            else:
                for thread in self.thread_list[:-clip_index]:
                    file_name = thread.name
                    shutil.move(
                        os.path.join(self.img_dir, file_name),
                        os.path.join(self.move_folder, file_name),
                    )



import os
import shutil

import torch

from pepedp.embedding.embedding_class import ImgToEmbedding
from pepedp.scripts.utils.distance import euclid_dist
from pepeline import read, ImgColor, ImgFormat


def create_embedd(
    img_folder: str,
    embedder: ImgToEmbedding = ImgToEmbedding(),
):
    embedded = {}
    for img_name in os.listdir(img_folder):
        img = read(os.path.join(img_folder, img_name), ImgColor.RGB, ImgFormat.F32)
        embedded[img_name] = embedder(img).detach().cpu()
    return embedded


def filtered_pairs(
    embeddings, dist_func=euclid_dist, threshold: float = 1.5, device_str: str = None
):
    names = list(embeddings.keys())

    tensor_list = [embeddings[name] for name in names]
    E = torch.stack(tensor_list, dim=0)
    E = E.view(E.size(0), -1)

    device = (
        torch.device(device_str)
        if device_str
        else (
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        )
    )
    E = E.to(device)

    N = E.size(0)
    filtered_pairs = []

    for i in range(N):
        if i % 100 == 0 or i == N - 1:
            print(f"{i + 1} x {N}")

        anchor = E[i].unsqueeze(0)  # [1, D]
        compare_to = E[i + 1 :]  # [N - i - 1, D]
        if compare_to.size(0) == 0:
            continue

        dists = dist_func(anchor, compare_to).squeeze(0)  # [N - i - 1]
        mask = dists < threshold
        j_indices = torch.nonzero(mask).squeeze(1) + (i + 1)

        for idx, dist_val in zip(j_indices.tolist(), dists[mask].tolist()):
            if dist_val <= 0.01:
                print(dist_val, idx, names[idx], names[i])
            filtered_pairs.append((i, idx, dist_val))

    return {
        "names": names,
        "filtered_pairs": filtered_pairs,
    }


def move_duplicate_files(
    duplicates_dict: str, src_dir: str = "", dst_dir: str = ""
) -> None:
    os.makedirs(dst_dir, exist_ok=True)

    names = duplicates_dict["names"]
    duplicates = duplicates_dict["filtered_pairs"]

    seen_indices = set()
    for i, j, _ in duplicates:
        seen_indices.add(i)
        seen_indices.add(j)

    moved_count = 0
    for idx in seen_indices:
        filename = names[idx]
        src_path = os.path.join(src_dir, filename)
        dst_path = os.path.join(dst_dir, filename)

        if not os.path.exists(src_path):
            print(f"[!] File not detect: {src_path}")
            continue

        if os.path.exists(dst_path):
            continue

        os.makedirs(os.path.dirname(dst_path), exist_ok=True)

        shutil.move(src_path, dst_path)
        moved_count += 1

    print(f"✅ Moved {moved_count} duplicate files to: {dst_dir}")



from pyiqa import create_metric

from pepedp.scripts.utils.objects import IQANode


class TopIQThread(IQANode):
    def __init__(
        self,
        img_dir,
        batch_size: int = 8,
        thread: float = 0.5,
        median_thread=0,
        move_folder: str | None = None,
    ):
        super().__init__(img_dir, batch_size, thread, median_thread, move_folder, None)
        self.model = create_metric("topiq_nr", device=self.device)

    def forward(self, images):
        return self.model(images)



from .aniqa import AnIQAThread
from .blocklines_iqa import BlockinessThread
from .hyper_iqa import HyperThread
from .topiq import TopIQThread
from .IC9600_iqa import IC9600Thread

__all__ = [
    "AnIQAThread",
    "HyperThread",
    "TopIQThread",
    "BlockinessThread",
    "IC9600Thread",
]



from abc import ABC, abstractmethod
import numpy as np


class BaseComplexity(ABC):
    @staticmethod
    @abstractmethod
    def type() -> str: ...

    @abstractmethod
    def get_tile_comp_score(
        self, image, complexity, y: int, x: int, tile_size: int
    ): ...

    @abstractmethod
    def __call__(self, img: np.ndarray): ...



import torch
from pepeline import resize, ResizesFilter, ResizesAlg
from pepedp.embedding.convnext import convnext_small, convnext_large
from pepedp.embedding.enum import EmbeddedModel
import torch.nn.functional as F


def enum_to_model(enum: EmbeddedModel):
    match enum:
        case EmbeddedModel.ConvNextS:
            return convnext_small()
        case EmbeddedModel.ConvNextL:
            return convnext_large()
        case EmbeddedModel.VITS:
            return torch.hub.load("facebookresearch/dinov2", "dinov2_vits14").eval()
        case EmbeddedModel.VITB:
            return torch.hub.load("facebookresearch/dinov2", "dinov2_vitb14").eval()
        case EmbeddedModel.VITL:
            return torch.hub.load("facebookresearch/dinov2", "dinov2_vitl14").eval()
        case EmbeddedModel.VITG:
            return torch.hub.load("facebookresearch/dinov2", "dinov2_vitg14").eval()


class ImgToEmbedding:
    def __init__(
        self,
        model: EmbeddedModel = EmbeddedModel.ConvNextS,
        amp: bool = True,
        scale: int = 4,
        device: str = "cuda",
    ):
        self.device = torch.device(device)
        self.scale = scale
        self.amp = amp
        self.model = enum_to_model(model).to(self.device)
        self.vit = model in [
            EmbeddedModel.VITS,
            EmbeddedModel.VITB,
            EmbeddedModel.VITL,
            EmbeddedModel.VITG,
        ]

    @staticmethod
    def check_img_size(x):
        b, c, h, w = x.shape
        mod_pad_h = (14 - h % 14) % 14
        mod_pad_w = (14 - w % 14) % 14
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), "reflect")

    def img_to_tensor(self, x):
        if self.vit:
            return self.check_img_size(
                torch.tensor(x.transpose((2, 0, 1)))[None, :, :, :].to(self.device)
            )
        return torch.tensor(x.transpose((2, 0, 1)))[None, :, :, :].to(self.device)

    @torch.inference_mode()
    def __call__(self, x):
        if self.scale > 1:
            h, w = x.shape[:2]
            x = resize(
                x,
                h // self.scale,
                w // self.scale,
                ResizesAlg.Conv(ResizesFilter.Bilinear),
                False,
            )
        with torch.amp.autocast(self.device.__str__(), torch.float16, self.amp):
            x = self.img_to_tensor(x)
            return self.model(x)



from enum import Enum
from pepedp.scripts.iqa import (
    HyperThread,
    AnIQAThread,
    TopIQThread,
    BlockinessThread,
    IC9600Thread,
)


class ThreadAlg(Enum):
    HIPERIQA = HyperThread
    ANIIQA = AnIQAThread
    TOPIQ = TopIQThread
    BLOCKINESS = BlockinessThread
    IC9600 = IC9600Thread



import numpy as np
import torch
import torch.fft

DEFAULT_BLOCK_SIZE = 8


try:
    """
    Author: Ziyang Hu
    Author's repo: https://github.com/zh217/torch-dct
    Licence: MIT
    """
    # PyTorch 1.7.0 and newer versions
    import torch.fft

    def dct1_rfft_impl(x):
        return torch.view_as_real(torch.fft.rfft(x, dim=1))

    def dct_fft_impl(v):
        return torch.view_as_real(torch.fft.fft(v, dim=1))

except ImportError:

    def dct_fft_impl(v):
        return torch.rfft(v, 1, onesided=False)


def dct(x, norm=None):
    """
    Author: Ziyang Hu
    Author's repo: https://github.com/zh217/torch-dct
    Licence: MIT

    Discrete Cosine Transform, Type II (a.k.a. the DCT)

    For the meaning of the parameter `norm`, see:
    https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.fftpack.dct.html

    :param x: the input signal
    :param norm: the normalization, None or 'ortho'
    :return: the DCT-II of the signal over the last dimension
    """
    x_shape = x.shape
    N = x_shape[-1]
    x = x.contiguous().view(-1, N)

    v = torch.cat([x[:, ::2], x[:, 1::2].flip([1])], dim=1)

    Vc = dct_fft_impl(v)

    k = -torch.arange(N, dtype=x.dtype, device=x.device)[None, :] * np.pi / (2 * N)
    width_r = torch.cos(k)
    width_i = torch.sin(k)

    V = Vc[:, :, 0] * width_r - Vc[:, :, 1] * width_i

    if norm == "ortho":
        V[:, 0] /= np.sqrt(N) * 2
        V[:, 1:] /= np.sqrt(N / 2) * 2

    V = 2 * V.view(*x_shape)

    return V


def dct_2d(x, norm=None):
    """
    Author: Ziyang Hu
    Author's repo: https://github.com/zh217/torch-dct
    Licence: MIT

    2-dimentional Discrete Cosine Transform, Type II (a.k.a. the DCT)

    For the meaning of the parameter `norm`, see:
    https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.fftpack.dct.html

    :param x: the input signal
    :param norm: the normalization, None or 'ortho'
    :return: the DCT-II of the signal over the last 2 dimensions
    """
    X1 = dct(x, norm=norm)
    X2 = dct(X1.transpose(-1, -2), norm=norm)
    return X2.transpose(-1, -2)


def calc_margin(
    height: int,
    width: int,
    block_size: int = DEFAULT_BLOCK_SIZE,
) -> tuple[int, int, int, int]:
    """
    Calculate margins for DCT processing.
    """
    height_margin: int = height % block_size
    width_margin: int = width % block_size
    cal_height: int = height - (
        height_margin if height_margin >= 4 else height_margin + block_size
    )
    cal_width: int = width - (
        width_margin if width_margin >= 4 else width_margin + block_size
    )
    height_margin = (height_margin + block_size) if height_margin < 4 else height_margin
    width_margin = (width_margin + block_size) if width_margin < 4 else width_margin
    return cal_height, cal_width, height_margin, width_margin


def calc_v_torch(
    dct_img: torch.Tensor,
    height_block_num: int,
    width_block_num: int,
    block_size: int = DEFAULT_BLOCK_SIZE,
) -> torch.Tensor:
    """Calculate averaged V values from a batched DCT image.

    This function computes the V value for each block in the DCT image by extracting the
    central pixel and its four neighbors (left, right, top, bottom) from each block.
    The V value for each pixel is defined as:
      V = sqrt((b_val + c_val - 2 * a)² + (d_val + e_val - 2 * a)²)
    where:
      - a is the central pixel value,
      - b_val and c_val are the left and right neighbor values,
      - d_val and e_val are the top and bottom neighbor values.
    The computed V values are then averaged over all blocks for each image in the batch.

    The input tensor should have shape (B, H, W) or (B, 1, H, W), where B is the batch size.
    The block indices are computed using offsets derived from the block size.

    Args:
        dct_img: A tensor representing the DCT image with shape (B, H, W) or (B, 1, H, W).
        height_block_num: Number of blocks along the height of the image.
        width_block_num: Number of blocks along the width of the image.
        block_size: Size of each block. Defaults to DEFAULT_BLOCK_SIZE.

    Returns:
        A tensor of shape (B,) containing the averaged V values for each image in the batch.

    Raises:
        TypeError: If dct_img is not a torch.Tensor.
        ValueError: If dct_img has an unsupported number of dimensions, or if block dimensions are invalid.
    """
    if not isinstance(dct_img, torch.Tensor):
        raise TypeError(f"dct_img must be a torch.Tensor, got {type(dct_img)}")

    if dct_img.dim() not in (3, 4):
        raise ValueError(
            f"dct_img must have 3 or 4 dimensions (got {dct_img.dim()}). Expected (B, H, W) or (B, 1, H, W)."
        )

    if block_size <= 0:
        raise ValueError(f"block_size must be positive, got {block_size}")
    if height_block_num <= 2 or width_block_num <= 2:
        raise ValueError(
            f"height_block_num and width_block_num must be greater than 2 (got {height_block_num} and {width_block_num})."
        )

    if dct_img.dim() == 4:
        dct_img = dct_img.squeeze(1)

    batch_size = dct_img.size(0)

    # compute number of offsets (blocks used for indexing)
    num_h = height_block_num - 3
    num_w = width_block_num - 3

    device = dct_img.device

    # compute starting offsets for each block
    # Offsets: block_size + (index * block_size) for index 1 to height_block_num-3 (inclusive)
    h_offsets = (
        block_size + torch.arange(1, height_block_num - 2, device=device) * block_size
    )  # shape: (num_h,)

    w_offsets = (
        block_size + torch.arange(1, width_block_num - 2, device=device) * block_size
    )  # shape: (num_w,)

    # build the row and column indices for blocks
    # r: shape (num_h, 1, block_size, 1), c: shape (1, num_w, 1, block_size)
    r = h_offsets.view(num_h, 1, 1, 1) + torch.arange(block_size, device=device).view(
        1, 1, block_size, 1
    )
    c = w_offsets.view(1, num_w, 1, 1) + torch.arange(block_size, device=device).view(
        1, 1, 1, block_size
    )

    # expand indices to full grid shape (num_h, num_w, block_size, block_size) and cast to int
    r = r.expand(num_h, num_w, block_size, block_size).to(torch.int)
    c = c.expand(num_h, num_w, block_size, block_size).to(torch.int)

    # create a batch index tensor of shape (B, 1, 1, 1, 1) for advanced indexing
    batch_idx = torch.arange(batch_size, device=device).view(batch_size, 1, 1, 1, 1)

    # expand r and c to include the batch dimension: shape becomes (B, num_h, num_w, block_size, block_size)
    r_exp = r.unsqueeze(0).expand(batch_size, num_h, num_w, block_size, block_size)
    c_exp = c.unsqueeze(0).expand(batch_size, num_h, num_w, block_size, block_size)

    # Extract the central value (a) and its four neighbors:
    # left (b_val), right (c_val), top (d_val), and bottom (e_val)
    a = dct_img[batch_idx, r_exp, c_exp]
    b_val = dct_img[batch_idx, r_exp, c_exp - block_size]
    c_val = dct_img[batch_idx, r_exp, c_exp + block_size]
    d_val = dct_img[batch_idx, r_exp - block_size, c_exp]
    e_val = dct_img[batch_idx, r_exp + block_size, c_exp]

    # compute V for each block and pixel within the block
    V = torch.sqrt((b_val + c_val - 2 * a) ** 2 + (d_val + e_val - 2 * a) ** 2)

    # average V over all blocks for each image
    normalization = (height_block_num - 2) * (width_block_num - 2)
    V_average = V.sum(dim=(1, 2)) / normalization

    return V_average


def blockwise_dct(
    gray_imgs: torch.Tensor,
    height_block_num: int,
    width_block_num: int,
    block_size: int = DEFAULT_BLOCK_SIZE,
) -> torch.Tensor:
    """Compute the DCT of an image block-wise using batched processing.

    This function divides the image into non-overlapping blocks of size block_size x block_size,
    applies a batched DCT transform (using dct2, which accepts batched input), and reconstructs the
    DCT image from the transformed blocks.

    Args:
        gray_imgs: Input image tensor with shape (H, W).
        dct: A DCT object with a method dct2 that accepts batched input.
        height_block_num: Number of blocks along the height.
        width_block_num: Number of blocks along the width.

    Returns:
        A tensor containing the DCT coefficients of the image blocks, arranged in the original block layout.
    """
    assert gray_imgs.dim() == 4, "Input tensor must have shape (B, 1, H, W)."
    batch_size, channel_size, *_ = gray_imgs.shape
    assert channel_size == 1, "Input tensor must have shape (B, 1, H, W)."

    if (
        gray_imgs.shape[-2] < height_block_num * block_size
        or gray_imgs.shape[-1] < width_block_num * block_size
    ):
        raise ValueError(
            f"Invalid image dimensions. Image must be at least {height_block_num * block_size} x {width_block_num * block_size} because of number of blocks and block size."
        )

    # divide the image into blocks of shape (height_block_num, width_block_num, block_size, block_size).
    blocks = gray_imgs.unfold(
        -2,
        block_size,
        block_size,
    ).unfold(
        -1,
        block_size,
        block_size,
    )
    blocks = blocks.contiguous().view(batch_size, -1, block_size, block_size)

    # apply the batched DCT transform to all blocks at once.
    dct_blocks_flat: torch.Tensor = dct_2d(blocks, norm="ortho")

    dct_blocks = dct_blocks_flat.view(
        batch_size,
        height_block_num,
        width_block_num,
        block_size,
        block_size,
    )
    dct_blocks = dct_blocks.permute(0, 1, 3, 2, 4).contiguous()
    dct_blocks = dct_blocks.view(
        batch_size,
        height_block_num * block_size,
        width_block_num * block_size,
    )
    return dct_blocks


def calculate_image_blockiness(
    rgb_images: torch.Tensor,
    block_size: int = DEFAULT_BLOCK_SIZE,
):
    """Calculate the blockiness metric for a batch of grayscale images.

    The blockiness is computed by comparing the blockwise DCT of the original image and an
    offset version of the image. The offset image is created by shifting the original image by
    4 pixels in both spatial dimensions.

    Args:
        rgb_images: A 4D tensor with shape (B, 1, H, W) representing a batch of grayscale images.
        block_size: The size of each block used for the DCT computation. Defaults to DEFAULT_BLOCK_SIZE.

    Returns:
        A tensor containing the summed blockiness scores per image.

    Raises:
        TypeError: If gray_images is not a torch.Tensor.
        ValueError: If gray_images does not have 4 dimensions or if its spatial dimensions
                    are too small for offset calculation.
        RuntimeError: If any of the intermediate calculations fail.
    """
    gray_images = rgb_to_grayscale(rgb_images)
    if not isinstance(gray_images, torch.Tensor):
        raise TypeError("gray_images must be a torch.Tensor.")
    if gray_images.dim() != 4:
        raise ValueError("Input tensor must have shape (B, 1, H, W).")

    # Extract height and width from the spatial dimensions.
    height, width = gray_images.shape[-2:]
    if height < 4 or width < 4:
        raise ValueError(
            "Image height and width must be at least 4 pixels for offset calculation."
        )

    # Calculate the valid height and width margins.
    cal_height, cal_width, _, _ = calc_margin(height=height, width=width)

    height_block_num, width_block_num = (
        cal_height // block_size,
        cal_width // block_size,
    )

    gray_tensor_cut = gray_images[..., :cal_height, :cal_width]
    gray_offset = torch.zeros_like(gray_images)
    gray_offset[..., :-4, :-4] = gray_images[..., 4:, 4:]
    gray_offset = gray_offset[..., :cal_height, :cal_width]

    dct_imgs = blockwise_dct(
        gray_imgs=gray_tensor_cut,
        height_block_num=height_block_num,
        width_block_num=width_block_num,
        block_size=block_size,
    )

    dct_offset_imgs = blockwise_dct(
        gray_imgs=gray_offset,
        height_block_num=height_block_num,
        width_block_num=width_block_num,
        block_size=block_size,
    )
    v_average = calc_v_torch(
        dct_img=dct_imgs,
        height_block_num=height_block_num,
        width_block_num=width_block_num,
        block_size=block_size,
    )
    v_offset_average = calc_v_torch(
        dct_img=dct_offset_imgs,
        height_block_num=height_block_num,
        width_block_num=width_block_num,
        block_size=block_size,
    )
    epsilon = 1e-8
    d = torch.abs(v_offset_average - v_average) / (v_average + epsilon)
    d_sum = torch.sum(d, dim=(1, 2))
    return d_sum


def rgb_to_grayscale(tensor):
    # define luminance coefficients
    weights = torch.tensor([0.2989, 0.5870, 0.1140], device=tensor.device).view(
        1, 3, 1, 1
    )

    # apply weighted sum across the channel dimension
    # (B, 1, H, W)
    grayscale = (tensor * weights).sum(dim=1, keepdim=True)
    return grayscale



import os
from pepeline import (
    read,
    save,
    best_tile,
    ImgColor,
    ImgFormat,
    ResizesFilter,
    ResizesAlg,
    resize,
)
from tqdm.contrib.concurrent import process_map, thread_map
from tqdm import tqdm
from pepedp.enum import ProcessType
from pepedp.scripts.utils.complexity.laplacian import LaplacianComplexity
from pepedp.scripts.utils.complexity.object import BaseComplexity


class BestTile:
    """
    Class for processing images to find and save the best tile with the highest Laplacian intensity.

    Attributes:
    - `in_folder` (str): Input folder containing images.
    - `out_folder` (str): Output folder to save processed images.
    - `tile_size` (int): Size of the tile to extract. Default is 1024.
    - `process_type` (str): Type of processing ('thread', 'process', or 'sequential'). Default is 'thread'.
    - `all_images` (list): List of all image filenames in the input folder.

    Methods:
    - `process(img_name: str)`: Process a single image to find and save the best tile.
    - `run()`: Run the processing on all images using the specified processing type.
    """

    def __init__(
        self,
        in_folder: str,
        out_folder: str,
        tile_size: int = 512,
        process_type: ProcessType = ProcessType.THREAD,
        scale: int = 1,
        dynamic_n_tiles: bool = True,
        laplacian_thread: float = 0,
        image_gray: bool = False,
        func: BaseComplexity = LaplacianComplexity(),
    ):
        """
        Initialize the BestTile class.
        """
        self.scale = scale
        self.in_folder = in_folder
        self.out_folder = out_folder
        os.makedirs(out_folder, exist_ok=True)
        self.out_list = os.listdir(out_folder)
        self.tile_size = tile_size
        self.all_images = os.listdir(in_folder)
        self.process_type = process_type
        self.dynamic_n_tiles = dynamic_n_tiles

        self.laplacian_thread = laplacian_thread
        self.image_gray = image_gray
        self.func = func
        if func.type() == "IC9600":
            self.process_type = ProcessType.FOR

    def save_result(self, img, img_name) -> None:
        save(img, os.path.join(self.out_folder, img_name))

    def get_tile(self, img, complexity):
        if self.func.type() != "Laplacian":
            left_up_cord = best_tile(complexity[0], self.tile_size // 8)
        elif self.scale > 1:
            img_shape = complexity.shape
            complexity_r = resize(
                complexity,
                img_shape[0] // self.scale,
                img_shape[1] // self.scale,
                ResizesAlg.Conv(ResizesFilter.Bilinear),
                False,
            ).squeeze()
            left_up_cord = best_tile(complexity_r, self.tile_size // self.scale)
            left_up_cord = [index * self.scale for index in left_up_cord]
        else:
            left_up_cord = best_tile(complexity, self.tile_size)
        img, laplacian, score = self.func.get_tile_comp_score(
            img, complexity, left_up_cord[0], left_up_cord[1], self.tile_size
        )
        return img, laplacian, score

    def read_img(self, img_name):
        image = read(
            os.path.join(self.in_folder, img_name),
            ImgColor.GRAY if self.image_gray else ImgColor.RGB,
            ImgFormat.F32,
        )
        return image

    def process(self, img_name: str):
        """
        Process a single image to find and save the best tile.

        Args:
        - `img_name` (str): Name of the image file to process.
        """
        try:
            if img_name.split(".")[0] + ".png" in self.out_list:
                return
            img = self.read_img(img_name)
            img_shape = img.shape
            if img_shape[0] < self.tile_size or img_shape[1] < self.tile_size:
                return
            result_name = ".".join(img_name.split(".")[:-1]) + ".png"
            if img_shape[0] == self.tile_size and img_shape[1] == self.tile_size:
                self.save_result(img, result_name)
                return
            complexity = self.func(img)
            if (
                img_shape[0] * img_shape[1] > self.tile_size**2 * 4
                and self.dynamic_n_tiles
            ):
                for i in range(
                    (img_shape[0] * img_shape[1]) // (self.tile_size**2 * 2)
                ):
                    tile, laplacian_abs, score = self.get_tile(img, complexity)
                    if self.laplacian_thread:
                        if score < self.laplacian_thread:
                            break
                    self.save_result(
                        tile, ".".join(img_name.split(".")[:-1]) + f"_{i}" + ".png"
                    )
            else:
                tile, laplacian_abs, score = self.get_tile(img, complexity)
                if self.laplacian_thread:
                    if score < self.laplacian_thread:
                        return
                self.save_result(tile, result_name)
        except Exception as e:
            print(img_name, "\n", e)

    def run(self):
        """
        Run the processing on all images using the specified processing type.
        """
        if self.process_type == ProcessType.THREAD:
            thread_map(self.process, self.all_images)
        elif self.process_type == ProcessType.PROCESS:
            process_map(self.process, self.all_images)
        else:
            for img_name in tqdm(self.all_images):
                self.process(img_name)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models import resnet18, ResNet18_Weights


class slam(nn.Module):
    def __init__(self, spatial_dim):
        super(slam, self).__init__()
        self.spatial_dim = spatial_dim
        self.linear = nn.Sequential(
            nn.Linear(spatial_dim**2, 512), nn.ReLU(), nn.Linear(512, 1), nn.Sigmoid()
        )

    def forward(self, feature):
        n, c, h, w = feature.shape
        if h != self.spatial_dim:
            x = F.interpolate(
                feature,
                size=(self.spatial_dim, self.spatial_dim),
                mode="bilinear",
                align_corners=True,
            )
        else:
            x = feature

        x = x.view(n, c, -1)
        x = self.linear(x)
        x = x.unsqueeze(dim=3)
        out = x.expand_as(feature) * feature

        return out


class to_map(nn.Module):
    def __init__(self, channels):
        super(to_map, self).__init__()
        self.to_map = nn.Sequential(
            nn.Conv2d(in_channels=channels, out_channels=1, kernel_size=1, stride=1),
            nn.Sigmoid(),
        )

    def forward(self, feature):
        return self.to_map(feature)


class conv_bn_relu(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size=3, padding=1, stride=1):
        super(conv_bn_relu, self).__init__()
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            padding=padding,
            stride=stride,
        )
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU()

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class up_conv_bn_relu(nn.Module):
    def __init__(
        self, scale, in_channels, out_channels=64, kernal_size=1, padding=0, stride=1
    ):
        super(up_conv_bn_relu, self).__init__()
        self.upSample = (
            nn.Upsample(scale_factor=scale, mode="bilinear", align_corners=True)
            if scale != 1
            else nn.Identity()
        )
        self.conv = nn.Conv2d(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernal_size,
            stride=stride,
            padding=padding,
        )
        self.bn = nn.BatchNorm2d(num_features=out_channels)
        self.act = nn.ReLU()

    def forward(self, x):
        x = self.upSample(x)
        x = self.conv(x)
        x = self.bn(x)
        x = self.act(x)
        return x


class ICNet(nn.Module):
    def __init__(self, is_pretrain=True, size1=512, size2=256):
        super(ICNet, self).__init__()
        resnet18Pretrained1 = resnet18(weights=ResNet18_Weights.IMAGENET1K_V1)
        resnet18Pretrained2 = resnet18(weights=ResNet18_Weights.IMAGENET1K_V1)

        self.size1 = size1
        self.size2 = size2

        ## detail branch
        self.b1_1 = nn.Sequential(*list(resnet18Pretrained1.children())[:5])
        self.b1_1_slam = slam(32)

        self.b1_2 = list(resnet18Pretrained1.children())[5]
        self.b1_2_slam = slam(32)

        ## context branch
        self.b2_1 = nn.Sequential(*list(resnet18Pretrained2.children())[:5])
        self.b2_1_slam = slam(32)

        self.b2_2 = list(resnet18Pretrained2.children())[5]
        self.b2_2_slam = slam(32)

        self.b2_3 = list(resnet18Pretrained2.children())[6]
        self.b2_3_slam = slam(16)

        self.b2_4 = list(resnet18Pretrained2.children())[7]
        self.b2_4_slam = slam(8)

        self.up1 = up_conv_bn_relu(1, in_channels=128, out_channels=256)
        self.up2 = up_conv_bn_relu(8, in_channels=512, out_channels=256)

        ## map prediction head
        self.to_map_f = conv_bn_relu(256 * 2, 256 * 2)
        self.to_map_f_slam = slam(32)
        self.to_map = to_map(256 * 2)

        ## score prediction head
        self.to_score_f = conv_bn_relu(256 * 2, 256 * 2)
        self.to_score_f_slam = slam(32)
        self.head = nn.Sequential(
            nn.Linear(256 * 2, 512), nn.ReLU(), nn.Linear(512, 1), nn.Sigmoid()
        )
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

    def check_img_size(self, x):
        b, c, h, w = x.shape
        scaled_size = 64
        mod_pad_h = (scaled_size - h % scaled_size) % scaled_size
        mod_pad_w = (scaled_size - w % scaled_size) % scaled_size
        return F.pad(x, (0, mod_pad_w, 0, mod_pad_h), "reflect")

    def get_only_score(self, x1):
        x_cat = self.body_forward(x1)
        return self.score(x_cat)

    def score(self, x_cat):
        score_feature = self.to_score_f_slam(self.to_score_f(x_cat))
        score_feature = self.avgpool(score_feature)
        score_feature = score_feature.squeeze()
        score = self.head(score_feature)
        score = score.squeeze()
        return score

    def body_forward(self, x1):
        b, c, h, w = x1.shape
        x1 = F.pad(
            x1, pad=(32, 32, 32, 32), mode="reflect"
        )  # (left, right, top, bottom)
        # assert(x1.shape[2] == x1.shape[3] == self.size1)
        x1 = self.check_img_size(x1)
        x2 = F.interpolate(x1, scale_factor=0.5, mode="bilinear", align_corners=True)

        x1 = self.b1_2_slam(self.b1_2(self.b1_1_slam(self.b1_1(x1))))
        x2 = self.b2_2_slam(self.b2_2(self.b2_1_slam(self.b2_1(x2))))
        x2 = self.b2_4_slam(self.b2_4(self.b2_3_slam(self.b2_3(x2))))

        x1 = self.up1(x1)
        x2 = self.up2(x2)
        return torch.cat((x1, x2), dim=1)[:, :, 4:-4, 4:-4][:, :, : h // 8, : w // 8]

    def forward(self, x1):
        x_cat = self.body_forward(x1)
        cly_map = self.to_map(self.to_map_f_slam(self.to_map_f(x_cat)))
        return x_cat, cly_map


def ic9600():
    state = torch.hub.load_state_dict_from_url(
        url="https://github.com/umzi2/Dataset_Preprocessing/releases/download/IC9600_duplicate/ic9600.pth",
        map_location="cpu",
        weights_only=True,
    )
    model = ICNet().eval()
    model.load_state_dict(state)
    return model



import torch
from torchvision.transforms import transforms
from pepedp.scripts.utils.objects import IQANode


class AnIQAThread(IQANode):
    def __init__(
        self,
        img_dir,
        batch_size: int = 8,
        thread: float = 0.5,
        median_thread=0,
        move_folder: str | None = None,
    ):
        compose = transforms.Compose(
            [
                transforms.Normalize(
                    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                ),
            ]
        )
        super().__init__(
            img_dir, batch_size, thread, median_thread, move_folder, compose
        )
        self.model = torch.hub.load(
            repo_or_dir="miccunifi/ARNIQA",
            source="github",
            model="ARNIQA",
            regressor_dataset="kadid10k",
        )
        self.model.eval().to(self.device)

    def forward(self, images):
        _, _, h, w = images.size()
        images_ds = transforms.Resize((h // 2, w // 2)).to(images)(images)
        with torch.inference_mode():
            iqa = self.model(
                images, images_ds, return_embedding=False, scale_score=True
            )
        return iqa



import cv2
import numpy as np

from pepedp.scripts.utils.complexity.object import BaseComplexity


class LaplacianComplexity(BaseComplexity):
    def __init__(self, median_blur: int = 1):
        super().__init__()
        self.median_blur = median_blur if median_blur % 2 == 1 else median_blur + 1

    @staticmethod
    def image_to_gray(image: np.ndarray):
        image = image.squeeze()
        if image.ndim == 3:
            return cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)
        return image

    def median_laplacian(self, image):
        if self.median_blur <= 1:
            return image
        elif self.median_blur <= 5:
            image = cv2.medianBlur(image, self.median_blur)
        else:
            image = (
                cv2.medianBlur((image * 255).astype(np.uint8), self.median_blur).astype(
                    np.float32
                )
                / 255
            )
        return image

    @staticmethod
    def type():
        return "Laplacian"

    @staticmethod
    def get_tile_comp_score(image, complexity, y, x, tile_size):
        img_tile = image[
            y : y + tile_size,
            x : x + tile_size,
        ]
        score = np.mean(
            complexity[
                y : y + tile_size,
                x : x + tile_size,
            ]
        )
        complexity[
            y : y + tile_size,
            x : x + tile_size,
        ] = -1.0
        return img_tile, complexity, score

    def __call__(self, img):
        img = self.image_to_gray(img)
        img = self.median_laplacian(img)

        return np.abs(cv2.Laplacian(img, -1))



import os

import cv2
import numpy as np

from tqdm import tqdm

from pepedp.embedding.embedding_class import ImgToEmbedding
from pepedp.scripts.utils.distance import euclid_dist


class VideoToFrame:
    def __init__(
        self,
        embedder: ImgToEmbedding = ImgToEmbedding(),
        thread: float = 0.3,
        distance_fn=euclid_dist,
    ):
        self.embedder = embedder
        self.thread = thread
        self.distance_func = distance_fn

    def __call__(self, video_path, out_path):
        os.makedirs(out_path, exist_ok=True)
        capture = cv2.VideoCapture(video_path)
        total_frames = int(capture.get(cv2.CAP_PROP_FRAME_COUNT))
        ref = None
        n = 0
        with tqdm(total=total_frames) as pbar:
            while capture.isOpened():
                ret, frame = capture.read()
                if not ret:
                    break
                if ref is None:
                    ref = self.embedder(
                        cv2.cvtColor(frame, cv2.COLOR_BGR2RGB).astype(np.float32)
                        / 255.0
                    )
                else:
                    temp_embedd = self.embedder(
                        cv2.cvtColor(frame, cv2.COLOR_BGR2RGB).astype(np.float32)
                        / 255.0
                    )
                    if self.distance_func(ref, temp_embedd).squ > self.thread:
                        cv2.imwrite(os.path.join(out_path, f"frame_{n}.png"), frame)
                        ref = temp_embedd
                n += 1
                pbar.update(1)

        capture.release()



import cv2
import numpy as np
import torch

from pepedp.scripts.archs.ICNet import ic9600
from pepedp.scripts.utils.complexity.object import BaseComplexity


class IC9600Complexity(BaseComplexity):
    def __init__(self, device: str = "cuda"):
        super().__init__()
        self.device = torch.device(device)
        self.arch = ic9600().to(self.device)

    @staticmethod
    def image_to_tensor(image: np.ndarray):
        image = image.squeeze()
        if image.ndim == 2:
            return torch.tensor(
                cv2.cvtColor(image, cv2.COLOR_GRAY2RGB).transpose((2, 0, 1))
            )[None, :, :, :]
        return torch.tensor(image.transpose((2, 0, 1)))[None, :, :, :]

    @staticmethod
    def type():
        return "IC9600"

    # @staticmethod
    @torch.inference_mode()
    def get_tile_comp_score(self, image, complexity, y, x, tile_size):
        img_tile = image[
            y * 8 : y * 8 + tile_size,
            x * 8 : x * 8 + tile_size,
        ]
        score = (
            self.arch.score(
                complexity[1][
                    :,
                    :,
                    y : y + tile_size // 8,
                    x : x + tile_size // 8,
                ]
            )
            .detach()
            .cpu()
            .item()
        )
        complexity[0][
            y : y + tile_size // 8,
            x : x + tile_size // 8,
        ] = -1.0
        return img_tile, complexity, score

    @torch.inference_mode()
    def __call__(self, img):
        img = self.image_to_tensor(img).to(self.device)
        x_cat, cly_map = self.arch(img)
        return cly_map.detach().cpu().squeeze().numpy(), x_cat



import torch
import torch.nn.functional as F


def cosine_dist(emb1: torch.Tensor, emb2: torch.Tensor) -> torch.Tensor | float:
    """
    Поддерживает сравнение:
    - [D] и [D]
    - [1, D] и [1, D]
    - [N, D] и [M, D] -> [N, M]
    """
    emb1 = emb1.unsqueeze(0) if emb1.ndim == 1 else emb1  # [D] → [1, D]
    emb2 = emb2.unsqueeze(0) if emb2.ndim == 1 else emb2

    emb1_norm = F.normalize(emb1, dim=-1)
    emb2_norm = F.normalize(emb2, dim=-1)

    sim = emb1_norm @ emb2_norm.T  # [N, D] @ [D, M] = [N, M]
    dist = 1 - sim

    # if dist.numel() == 1:
    #     return dist.item()
    return dist


def euclid_dist(
    emb1: torch.Tensor, emb2: torch.Tensor, p: float = 2.0
) -> torch.Tensor | float:
    """
    Поддерживает сравнение:
    - [D] и [D]
    - [1, D] и [1, D]
    - [N, D] и [M, D] -> [N, M]
    """
    emb1 = emb1.unsqueeze(0) if emb1.ndim == 1 else emb1  # [D] → [1, D]
    emb2 = emb2.unsqueeze(0) if emb2.ndim == 1 else emb2

    d = torch.cdist(emb1, emb2, p=p)
    # if d.numel() == 1:
    #     return d.item()
    return d
